#!/usr/bin/env python3
"""Global utility methods and classes."""

from distutils.util import strtobool


def to_bool(string):
    return bool(strtobool(str(string)))


def dict_intersect(d1, d2):
    return {
        k: dict_intersect(d1[k], d2[k]) if isinstance(d1[k], dict) else d1[k]
        for k in d1.keys() & d2.keys()
    }


class Singleton(type):
    """Singleton metaclass."""

    _instances = {}

    def __call__(cls, *args, **kwargs):

        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



"""Test Autostop class."""

import docker
import pytest

from dockertidy import autostop

pytest_plugins = [
    "dockertidy.test.fixtures.fixtures",
]


@pytest.fixture
def autostop_fixture(mocker):
    mocker.patch.object(
        autostop.AutoStop,
        "_get_docker_client",
        return_value=mocker.create_autospec(docker.APIClient)
    )

    stop = autostop.AutoStop()
    return stop


def test_stop_container(autostop_fixture, mocker):
    client = mocker.create_autospec(docker.APIClient)
    cid = "asdb"

    autostop_fixture._stop_container(client, cid)
    client.stop.assert_called_once_with(cid)


def test_build_container_matcher(autostop_fixture, mocker):
    prefixes = ["one_", "two_"]
    matcher = autostop_fixture._build_container_matcher(prefixes)

    assert matcher("one_container")
    assert matcher("two_container")
    assert not matcher("three_container")
    assert not matcher("one")


def test_has_been_running_since_true(autostop_fixture, container, later_time):
    assert autostop_fixture._has_been_running_since(container, later_time)


def test_has_been_running_since_false(autostop_fixture, container, earlier_time):
    assert not autostop_fixture._has_been_running_since(container, earlier_time)



#!/usr/bin/env python3
"""Custom input type parser."""

from argparse import ArgumentTypeError

import dateparser
import environs

env = environs.Env()


def timedelta_validator(value):
    """
    Return the dateparser string for a time in the past.

    :param value: a string containing a time format supported by
    mod:`dateparser`
    """
    if value is None:
        return None

    if not dateparser.parse(value):
        raise ArgumentTypeError(f"'{value}' is not a valid timedelta string")

    return value


def timedelta(value, dt_format=None):
    """
    Return the :class:`datetime.datetime.DateTime` for a time in the past.

    :param value: a string containing a time format supported by
    mod:`dateparser`
    """
    if value is None:
        return None

    timedelta = dateparser.parse(
        value, settings={
            "TO_TIMEZONE": "UTC",
            "RETURN_AS_TIMEZONE_AWARE": True
        }
    )

    if dt_format:
        timedelta = timedelta.strftime(dt_format)

    return timedelta


@env.parser_for("timedelta_validator")
def timedelta_parser(value):
    try:
        timedelta_validator(value)
        return value
    except ArgumentTypeError as e:
        raise environs.EnvError(e) from e



"""Test GarbageCollector class."""

import docker
import pytest
import requests

from dockertidy import garbage_collector

pytest_plugins = [
    "dockertidy.test.fixtures.fixtures",
]


@pytest.fixture
def gc(mocker):
    mocker.patch.object(
        garbage_collector.GarbageCollector,
        "_get_docker_client",
        return_value=mocker.create_autospec(docker.APIClient)
    )

    gc = garbage_collector.GarbageCollector()
    return gc


def test_is_running(gc, container, now):
    container["State"]["Running"] = True

    assert not gc._should_remove_container(container, now)


def test_is_ghost(gc, container, now):
    container["State"]["Ghost"] = True

    assert gc._should_remove_container(container, now)


def test_old_never_run(gc, container, now, earlier_time):
    container["Created"] = str(earlier_time)
    container["State"]["FinishedAt"] = gc.YEAR_ZERO

    assert gc._should_remove_container(container, now)


def test_not_old_never_run(gc, container, now, earlier_time):
    container["Created"] = str(now)
    container["State"]["FinishedAt"] = gc.YEAR_ZERO

    assert not gc._should_remove_container(container, now)


def test_old_stopped(gc, container, now):
    assert gc._should_remove_container(container, now)


def test_not_old(gc, container, now):
    container["State"]["FinishedAt"] = "2014-01-21T00:00:00Z"

    assert not gc._should_remove_container(container, now)


def test_cleanup_containers(gc, mocker, containers):
    client = mocker.create_autospec(docker.APIClient)
    client.containers.return_value = [
        {
            "Id": "abcd"
        },
        {
            "Id": "abbb"
        },
    ]
    client.inspect_container.side_effect = iter(containers)

    gc.config.config["gc"]["max_container_age"] = "0day"
    gc.docker = client
    gc.cleanup_containers()
    client.remove_container.assert_called_once_with(container="abcd", v=True)


def test_filter_excluded_containers(gc):
    containers = [
        {
            "Labels": {
                "toot": ""
            }
        },
        {
            "Labels": {
                "too": "lol"
            }
        },
        {
            "Labels": {
                "toots": "lol"
            }
        },
        {
            "Labels": {
                "foo": "bar"
            }
        },
        {
            "Labels": None
        },
    ]

    result = gc._filter_excluded_containers(containers)
    assert containers == list(result)

    gc.config.config["gc"]["exclude_container_labels"] = [
        gc.ExcludeLabel(key="too", value=None),
        gc.ExcludeLabel(key="foo", value=None),
    ]
    result = gc._filter_excluded_containers(containers)
    assert [containers[0], containers[2], containers[4]] == list(result)

    gc.config.config["gc"]["exclude_container_labels"] = [
        gc.ExcludeLabel(key="too*", value="lol"),
    ]
    result = gc._filter_excluded_containers(containers)
    assert [containers[0], containers[3], containers[4]] == list(result)


def test_cleanup_images(mocker, gc, containers):
    client = mocker.create_autospec(docker.APIClient)
    client._version = "1.21"
    client.images.return_value = images = [
        {
            "Id": "abcd"
        },
        {
            "Id": "abbb"
        },
    ]
    client.inspect_image.side_effect = iter(containers)

    gc.docker = client
    gc.config.config["gc"]["max_image_age"] = "0days"
    gc.cleanup_images(client)
    assert client.remove_image.mock_calls == [
        mocker.call(image=image["Id"]) for image in reversed(images)
    ]


def test_cleanup_volumes(mocker, gc):
    client = mocker.create_autospec(docker.APIClient)
    client.volumes.return_value = volumes = {
        "Volumes": [
            {
                "Mountpoint": "unused",
                "Labels": None,
                "Driver": "unused",
                "Name": u"one"
            },
            {
                "Mountpoint": "unused",
                "Labels": None,
                "Driver": "unused",
                "Name": u"two"
            },
        ],
        "Warnings": None,
    }

    gc.docker = client
    gc.cleanup_volumes()
    assert client.remove_volume.mock_calls == [
        mocker.call(name=volume["Name"]) for volume in reversed(volumes["Volumes"])
    ]


def test_filter_images_in_use(gc, images):
    image_tags_in_use = set([
        "user/one:latest",
        "user/foo:latest",
        "other-2:abc45",
        "2471708c19be:latest",
    ])
    expected = [
        {
            "RepoTags": ["<none>:<none>"],
            "Id": "babababababaabababab"
        },
        {
            "RepoTags": ["other-1:abcda"]
        },
        {
            "RepoTags": ["new_image:latest", "new_image:123"]
        },
    ]

    actual = gc._filter_images_in_use(images, image_tags_in_use)
    assert list(actual) == expected


def test_filter_images_in_use_by_id(mocker, gc, containers):
    client = mocker.create_autospec(docker.APIClient)
    client._version = "1.21"
    client.containers.return_value = [
        {
            "Id": "abcd",
            "ImageID": "1"
        },
        {
            "Id": "abbb",
            "ImageID": "2"
        },
    ]

    client.inspect_container.side_effect = iter(containers)
    client.images.return_value = [
        {
            "Id": "1",
            "Created": "2014-01-01T01:01:01Z"
        },
        {
            "Id": "2",
            "Created": "2014-01-01T01:01:01Z"
        },
        {
            "Id": "3",
            "Created": "2014-01-01T01:01:01Z"
        },
        {
            "Id": "4",
            "Created": "2014-01-01T01:01:01Z"
        },
        {
            "Id": "5",
            "Created": "2014-01-01T01:01:01Z"
        },
        {
            "Id": "6",
            "Created": "2014-01-01T01:01:01Z"
        },
    ]

    client.inspect_image.side_effect = lambda image: {
        "Id": image,
        "Created": "2014-01-01T01:01:01Z"
    }

    gc.docker = client
    gc.config.config["gc"]["max_image_age"] = "0days"
    gc.cleanup_images(set())
    assert client.remove_image.mock_calls == [
        mocker.call(image=id_) for id_ in ["6", "5", "4", "3"]
    ]


def test_filter_excluded_images(gc, images):
    exclude_set = set([
        "user/one:latest",
        "user/foo:latest",
        "other-2:abc45",
    ])
    expected = [
        {
            "RepoTags": ["<none>:<none>"],
            "Id": "2471708c19beabababab"
        },
        {
            "RepoTags": ["<none>:<none>"],
            "Id": "babababababaabababab"
        },
        {
            "RepoTags": ["other-1:abcda"]
        },
        {
            "RepoTags": ["new_image:latest", "new_image:123"]
        },
    ]

    actual = gc._filter_excluded_images(images, exclude_set)
    assert list(actual) == expected


def test_filter_excluded_images_advanced(gc, images):
    exclude_set = set([
        "user/one:*",
        "new_*:123",
        "other-1:abc*",
    ])
    expected = [
        {
            "RepoTags": ["<none>:<none>"],
            "Id": "2471708c19beabababab"
        },
        {
            "RepoTags": ["<none>:<none>"],
            "Id": "babababababaabababab"
        },
        {
            "RepoTags": ["other-2:abc45"]
        },
    ]

    actual = gc._filter_excluded_images(images, exclude_set)
    assert list(actual) == expected


def test_is_image_old(gc, image, now):
    assert gc._is_image_old(image, now)


def test_is_image_old_false(gc, image, later_time):
    assert not gc._is_image_old(image, later_time)


def test_remove_image_no_tags(mocker, gc, image, now):
    client = mocker.create_autospec(docker.APIClient)
    image_id = "abcd"
    image_summary = {"Id": image_id}
    client.inspect_image.return_value = image

    gc.docker = client
    gc._remove_image(image_summary, now)
    client.remove_image.assert_called_once_with(image=image_id)


def test_remove_image_new_image_not_removed(mocker, gc, image, later_time):
    client = mocker.create_autospec(docker.APIClient)
    image_id = "abcd"
    image_summary = {"Id": image_id}
    client.inspect_image.return_value = image

    gc.docker = client
    gc._remove_image(image_summary, later_time)
    assert not client.remove_image.mock_calls


def test_remove_image_with_tags(mocker, gc, image, now):
    client = mocker.create_autospec(docker.APIClient)
    image_id = "abcd"
    repo_tags = ["user/one:latest", "user/one:12345"]
    image_summary = {"Id": image_id, "RepoTags": repo_tags}
    client.inspect_image.return_value = image

    gc.docker = client
    gc._remove_image(image_summary, now)
    assert client.remove_image.mock_calls == [mocker.call(image=tag) for tag in repo_tags]


def test_api_call_success(mocker, gc):
    func = mocker.Mock()
    container = "abcd"
    result = gc._api_call(func, container=container)
    func.assert_called_once_with(container="abcd")

    assert result == func.return_value


def test_api_call_with_timeout(mocker, gc):
    func = mocker.Mock(side_effect=requests.exceptions.ReadTimeout("msg"), __name__="remove_image")
    image = "abcd"

    mock_log = mocker.patch.object(gc, "logger", autospec=True)
    gc._api_call(func, image=image)

    func.assert_called_once_with(image="abcd")
    mock_log.warning.assert_called_once_with("Failed to call remove_image " + "image=abcd msg")


def test_api_call_with_api_error(mocker, gc):
    func = mocker.Mock(
        side_effect=docker.errors.APIError(
            "Ooops",
            mocker.Mock(status_code=409, reason="Conflict", url="dummy"),
            explanation="failed"
        ),
        __name__="remove_image"
    )
    image = "abcd"

    mock_log = mocker.patch.object(gc, "logger", autospec=True)
    gc._api_call(func, image=image)

    func.assert_called_once_with(image="abcd")
    mock_log.warning.assert_called_once_with(
        "Error calling remove_image image=abcd "
        '409 Client Error for dummy: Conflict ("failed")'
    )


def test_get_all_containers(mocker, gc):
    client = mocker.create_autospec(docker.APIClient)
    count = 10
    client.containers.return_value = [mocker.Mock() for _ in range(count)]

    mock_log = mocker.patch.object(gc, "logger", autospec=True)

    gc.docker = client
    containers = gc._get_all_containers()
    assert containers == client.containers.return_value
    client.containers.assert_called_once_with(all=True)
    mock_log.info.assert_called_with("Found %s containers", count)


def test_get_all_images(mocker, gc):
    client = mocker.create_autospec(docker.APIClient)
    count = 7
    client.images.return_value = [mocker.Mock() for _ in range(count)]

    mock_log = mocker.patch.object(gc, "logger", autospec=True)

    gc.docker = client
    images = gc._get_all_images()
    assert images == client.images.return_value
    mock_log.info.assert_called_with("Found %s images", count)


def test_get_dangling_volumes(mocker, gc):
    client = mocker.create_autospec(docker.APIClient)
    count = 4
    client.volumes.return_value = {"Volumes": [mocker.Mock() for _ in range(count)]}

    mock_log = mocker.patch.object(gc, "logger", autospec=True)

    gc.docker = client
    volumes = gc._get_dangling_volumes()
    assert volumes == client.volumes.return_value["Volumes"]
    mock_log.info.assert_called_with("Found %s dangling volumes", count)


def test_build_exclude_set(gc):
    gc.config.config["gc"]["exclude_images"] = [
        "some_image:latest",
        "repo/foo:12345",
        "duplicate:latest",
    ]
    expected = set([
        "some_image:latest",
        "repo/foo:12345",
        "duplicate:latest",
    ])

    exclude_set = gc._build_exclude_set()
    assert exclude_set == expected


def test_format_exclude_labels(gc):
    gc.config.config["gc"]["exclude_container_labels"] = [
        "voo*",
        "doo=poo",
    ]
    expected = [
        gc.ExcludeLabel(key="voo*", value=None),
        gc.ExcludeLabel(key="doo", value="poo"),
    ]
    gc._format_exclude_labels()
    assert expected == gc.config.config["gc"]["exclude_container_labels"]


def test_build_exclude_set_empty(gc):
    gc.config.config["gc"]["exclude_images"] = []
    exclude_set = gc._build_exclude_set()
    assert exclude_set == set()


def test_get_docker_client(gc, mocker):
    assert isinstance(gc.docker, docker.APIClient)



#!/usr/bin/env python3
"""Entrypoint and CLI handler."""

import argparse

import dockertidy.exception
from dockertidy import __version__
from dockertidy.autostop import AutoStop
from dockertidy.config import SingleConfig
from dockertidy.garbage_collector import GarbageCollector
from dockertidy.logger import SingleLog
from dockertidy.parser import timedelta_validator


class DockerTidy:
    """Cli entrypoint to handle command arguments."""

    def __init__(self):
        self.log = SingleLog()
        self.logger = self.log.logger
        self.args = self._cli_args()
        self.config = self._get_config()
        self.gc = GarbageCollector()
        self.stop = AutoStop()
        self.run()

    def _cli_args(self):
        """
        Use argparse for parsing CLI arguments.

        :return: args objec
        """
        parser = argparse.ArgumentParser(description="keep docker hosts tidy")
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=None,
            dest="dry_run",
            help="only log actions, don't stop anything"
        )
        parser.add_argument(
            "-t",
            "--timeout",
            type=int,
            dest="http_timeout",
            metavar="HTTP_TIMEOUT",
            help="HTTP timeout in seconds for making docker API calls"
        )
        parser.add_argument(
            "-v", dest="logging.level", action="append_const", const=-1, help="increase log level"
        )
        parser.add_argument(
            "-q", dest="logging.level", action="append_const", const=1, help="decrease log level"
        )
        parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

        subparsers = parser.add_subparsers(dest="command", help="sub-command help")
        subparsers.required = True

        parser_gc = subparsers.add_parser("gc", help="run docker garbage collector")
        parser_gc.add_argument(
            "--max-container-age",
            type=timedelta_validator,
            dest="gc.max_container_age",
            metavar="MAX_CONTAINER_AGE",
            help="maximum age for a container, containers older than this age "
            "will be removed (dateparser value)"
        )
        parser_gc.add_argument(
            "--max-image-age",
            type=timedelta_validator,
            dest="gc.max_image_age",
            metavar="MAX_IMAGE_AGE",
            help="maxium age for an image, images older than this age will be "
            "removed (dateparser value)"
        )
        parser_gc.add_argument(
            "--dangling-volumes",
            action="store_true",
            dest="gc.dangling_volumes",
            help="dangling volumes will be removed"
        )
        parser_gc.add_argument(
            "--exclude-image",
            action="append",
            type=str,
            dest="gc.exclude_images",
            metavar="EXCLUDE_IMAGE",
            help="never remove images with this tag"
        )
        parser_gc.add_argument(
            "--exclude-container-label",
            action="append",
            type=str,
            dest="gc.exclude_container_labels",
            metavar="EXCLUDE_CONTAINER_LABEL",
            help="never remove containers with this label key "
            "or label key=value"
        )

        parser_stop = subparsers.add_parser(
            "stop", help="stop containers that have been running for too long"
        )
        parser_stop.add_argument(
            "--max-run-time",
            type=timedelta_validator,
            dest="stop.max_run_time",
            metavar="MAX_RUN_TIME",
            help="maximum time a container is allows to run (dateparser value)"
        )
        parser_stop.add_argument(
            "--prefix",
            action="append",
            type=str,
            dest="stop.prefix",
            metavar="PREFIX",
            help="only stop containers which match one of the prefix"
        )

        return parser.parse_args().__dict__

    def _get_config(self):
        try:
            config = SingleConfig(args=self.args)
        except dockertidy.exception.ConfigError as e:
            self.log.sysexit_with_message(e)

        try:
            self.log.set_level(config.config["logging"]["level"])
        except ValueError as e:
            self.log.sysexit_with_message(f"Can not set log level.\n{e!s}")

        self.logger.info(f"Using config file {config.config_file}")
        self.logger.debug(f"Config dump: {config.config}")

        return config

    def run(self):
        """Cli main method."""
        if self.config.config["command"] == "gc":
            self.gc.run()
        elif self.config.config["command"] == "stop":
            self.stop.run()


def main():
    DockerTidy()



#!/usr/bin/env python3
"""Remove unused docker containers and images."""

import fnmatch
from collections import namedtuple

import dateutil.parser
import docker
import docker.errors
import requests.exceptions

from dockertidy.config import SingleConfig
from dockertidy.logger import SingleLog
from dockertidy.parser import timedelta


class GarbageCollector:
    """Garbage collector object to handle cleanup tasks of container, images and volumes."""

    # This seems to be something docker uses for a null/zero date
    YEAR_ZERO = "0001-01-01T00:00:00Z"
    ExcludeLabel = namedtuple("ExcludeLabel", ["key", "value"])

    def __init__(self):
        self.config = SingleConfig()
        self.log = SingleLog()
        self.logger = SingleLog().logger
        self.docker = self._get_docker_client()

    def cleanup_containers(self):
        """Identify old containers and remove them."""
        config = self.config.config
        client = self.docker
        all_containers = self._get_all_containers()

        filtered_containers = self._filter_excluded_containers(all_containers)

        self.logger.info(
            "Removing containers older than '{}'".format(
                timedelta(config["gc"]["max_container_age"], dt_format="%Y-%m-%d, %H:%M:%S")
            )
        )

        for container_summary in reversed(list(filtered_containers)):
            container = self._api_call(
                client.inspect_container,
                container=container_summary["Id"],
            )
            if not container or not self._should_remove_container(
                container,
                timedelta(config["gc"]["max_container_age"]),
            ):
                continue

            self.logger.info(
                "Removing container {} {} {}".format(
                    container["Id"][:16],
                    container.get("Name", "").lstrip("/"), container["State"]["FinishedAt"]
                )
            )

            if not config["dry_run"]:
                self._api_call(
                    client.remove_container,
                    container=container["Id"],
                    v=True,
                )

    def _filter_excluded_containers(self, containers):
        config = self.config.config

        if not config["gc"]["exclude_container_labels"]:
            return containers

        def include_container(container):
            if self._should_exclude_container_with_labels(container):
                return False
            return True

        return filter(include_container, containers)

    def _should_exclude_container_with_labels(self, container):
        config = self.config.config

        if container["Labels"]:
            for exclude_label in config["gc"]["exclude_container_labels"]:
                if exclude_label.value:
                    matching_keys = fnmatch.filter(
                        container["Labels"].keys(),
                        exclude_label.key,
                    )
                    label_values_to_check = [
                        container["Labels"][matching_key] for matching_key in matching_keys
                    ]
                    if fnmatch.filter(label_values_to_check, exclude_label.value):
                        return True
                else:
                    if fnmatch.filter(container["Labels"].keys(), exclude_label.key):
                        return True
        return False

    def _should_remove_container(self, container, min_date):
        state = container.get("State", {})

        if state.get("Running"):
            return False

        if state.get("Ghost"):
            return True

        # Container was created, but never started
        if state.get("FinishedAt") == self.YEAR_ZERO:
            created_date = dateutil.parser.parse(container["Created"])
            return created_date < min_date

        finished_date = dateutil.parser.parse(state["FinishedAt"])
        return finished_date < min_date

    def _get_all_containers(self):
        client = self.docker
        self.logger.info("Getting all containers")
        containers = client.containers(all=True)
        self.logger.info("Found %s containers", len(containers))
        return containers

    def _get_all_images(self):
        client = self.docker
        self.logger.info("Getting all images")
        images = client.images()
        self.logger.info("Found %s images", len(images))
        return images

    def _get_dangling_volumes(self):
        client = self.docker
        self.logger.info("Getting dangling volumes")
        volumes = client.volumes({"dangling": True})["Volumes"] or []
        self.logger.info("Found %s dangling volumes", len(volumes))
        return volumes

    def cleanup_images(self, exclude_set):
        """Identify old images and remove them."""
        # re-fetch container list so that we don't  include removed containers
        client = self.docker
        config = self.config.config

        containers = self._get_all_containers()
        images = self._get_all_images()
        if docker.utils.compare_version("1.21", client._version) < 0:
            image_tags_in_use = {container["Image"] for container in containers}
            images = self._filter_images_in_use(images, image_tags_in_use)
        else:
            # ImageID field was added in 1.21
            image_ids_in_use = {container["ImageID"] for container in containers}
            images = self._filter_images_in_use_by_id(images, image_ids_in_use)
        images = self._filter_excluded_images(images, exclude_set)

        self.logger.info(
            "Removing images older than '{}'".format(
                timedelta(config["gc"]["max_image_age"], dt_format="%Y-%m-%d, %H:%M:%S")
            )
        )
        for image_summary in reversed(list(images)):
            self._remove_image(image_summary, timedelta(config["gc"]["max_image_age"]))

    def _filter_excluded_images(self, images, exclude_set):

        def include_image(image_summary):
            image_tags = image_summary.get("RepoTags")
            if self._no_image_tags(image_tags):
                return True
            for exclude_pattern in exclude_set:
                if fnmatch.filter(image_tags, exclude_pattern):
                    return False
            return True

        return filter(include_image, images)

    def _filter_images_in_use(self, images, image_tags_in_use):

        def get_tag_set(image_summary):
            image_tags = image_summary.get("RepoTags")
            if self._no_image_tags(image_tags):
                # The repr of the image Id used by client.containers()
                return {"{id}:latest".format(id=image_summary["Id"][:12])}
            return set(image_tags)

        def image_not_in_use(image_summary):
            return not get_tag_set(image_summary) & image_tags_in_use

        return filter(image_not_in_use, images)

    def _filter_images_in_use_by_id(self, images, image_ids_in_use):

        def image_not_in_use(image_summary):
            return image_summary["Id"] not in image_ids_in_use

        return filter(image_not_in_use, images)

    def _is_image_old(self, image, min_date):
        return dateutil.parser.parse(image["Created"]) < min_date

    def _no_image_tags(self, image_tags):
        return not image_tags or image_tags == ["<none>:<none>"]

    def _remove_image(self, image_summary, min_date):
        config = self.config.config
        client = self.docker
        image = self._api_call(client.inspect_image, image=image_summary["Id"])

        if not image or not self._is_image_old(image, min_date):
            return

        self.logger.info(f"Removing image {self._format_image(image, image_summary)}")
        if config["dry_run"]:
            return

        image_tags = image_summary.get("RepoTags")
        # If there are no tags, remove the id
        if self._no_image_tags(image_tags):
            self._api_call(client.remove_image, image=image_summary["Id"])
            return

        # Remove any repository tags so we don't  hit 409 Conflict
        for image_tag in image_tags:
            self._api_call(client.remove_image, image=image_tag)

    def _remove_volume(self, volume):
        config = self.config.config
        client = self.docker
        if not volume:
            return

        self.logger.info("Removing volume {name}".format(name=volume["Name"]))
        if config["dry_run"]:
            return

        self._api_call(client.remove_volume, name=volume["Name"])

    def cleanup_volumes(self):
        """Identify old volumes and remove them."""
        dangling_volumes = self._get_dangling_volumes()

        self.logger.info("Removing dangling volumes")
        for volume in reversed(dangling_volumes):
            self.logger.info("Removing dangling volume %s", volume["Name"])
            self._remove_volume(volume)

    def _api_call(self, func, **kwargs):
        try:
            return func(**kwargs)
        except requests.exceptions.Timeout as e:
            params = ",".join("%s=%s" % item for item in kwargs.items())  # noqa:UP031
            self.logger.warning(
                "Failed to call {name} {params} {msg}".format(
                    name=func.__name__, params=params, msg=str(e)
                )
            )
        except docker.errors.APIError as e:
            params = ",".join("%s=%s" % item for item in kwargs.items())  # noqa:UP031
            self.logger.warning(
                "Error calling {name} {params} {msg}".format(
                    name=func.__name__, params=params, msg=str(e)
                )
            )

    def _format_image(self, image, image_summary):

        def get_tags():
            tags = image_summary.get("RepoTags")
            if not tags or tags == ["<none>:<none>"]:
                return ""
            return ", ".join(tags)

        return "{id} {tags}".format(id=image["Id"][:16], tags=get_tags())

    def _build_exclude_set(self):
        config = self.config.config

        def is_image_tag(line):
            return line and not line.startswith("#")

        return set(config["gc"]["exclude_images"])

    def _format_exclude_labels(self):
        config = self.config.config
        exclude_labels = []

        for exclude_label_arg in config["gc"]["exclude_container_labels"]:
            split_exclude_label = exclude_label_arg.split("=", 1)
            exclude_label_key = split_exclude_label[0]
            exclude_label_value = split_exclude_label[1] if len(split_exclude_label) == 2 else None
            exclude_labels.append(
                self.ExcludeLabel(
                    key=exclude_label_key,
                    value=exclude_label_value,
                )
            )
        config["gc"]["exclude_container_labels"] = exclude_labels

    def _get_docker_client(self):
        config = self.config.config
        try:
            return docker.APIClient(version="auto", timeout=config["http_timeout"])
        except docker.errors.DockerException as e:
            self.log.sysexit_with_message(f"Can't create docker client\n{e}")

    def run(self):
        """Garbage collector main method."""
        self.logger.info("Start garbage collection")
        config = self.config.config
        self._format_exclude_labels()

        if config["gc"]["max_container_age"]:
            self.cleanup_containers()

        if config["gc"]["max_image_age"]:
            exclude_set = self._build_exclude_set()
            self.cleanup_images(exclude_set)

        if config["gc"]["dangling_volumes"]:
            self.cleanup_volumes()

        if (
            not config["gc"]["max_container_age"] and not config["gc"]["max_image_age"]
            and not config["gc"]["dangling_volumes"]
        ):
            self.logger.ing("Skipped, no arguments given")



#!/usr/bin/env python3
"""Custom exceptions."""


class TidyError(Exception):
    """Generic exception class for docker-tidy."""

    def __init__(self, msg, original_exception=""):
        super().__init__(f"{msg}\n{original_exception}")
        self.original_exception = original_exception


class ConfigError(TidyError):
    """Errors related to config file handling."""

    pass



#!/usr/bin/env python3
"""Stop long running docker iamges."""

import dateutil.parser
import docker
import docker.errors
import requests.exceptions

from dockertidy.config import SingleConfig
from dockertidy.logger import SingleLog
from dockertidy.parser import timedelta


class AutoStop:
    """Autostop object to handle long running containers."""

    def __init__(self):
        self.config = SingleConfig()
        self.log = SingleLog()
        self.logger = SingleLog().logger
        self.docker = self._get_docker_client()

    def stop_containers(self):
        """Identify long running containers and terminate them."""
        client = self.docker
        config = self.config.config

        max_run_time = timedelta(config["stop"]["max_run_time"])
        prefix = config["stop"]["prefix"]
        dry_run = config["dry_run"]

        matcher = self._build_container_matcher(prefix)

        self.logger.info(
            "Stopping containers older than '{}'".format(
                timedelta(config["stop"]["max_run_time"], dt_format="%Y-%m-%d, %H:%M:%S")
            )
        )
        for container_summary in client.containers():
            container = client.inspect_container(container_summary["Id"])
            name = container["Name"].lstrip("/")

            if (
                prefix and matcher(name) and self._has_been_running_since(container, max_run_time)
            ) or (not prefix and self._has_been_running_since(container, max_run_time)):
                self.logger.info(
                    "Stopping container {id} {name}: running since {started}".format(
                        id=container["Id"][:16],
                        name=name,
                        started=container["State"]["StartedAt"]
                    )
                )

                if not dry_run:
                    self._stop_container(client, container["Id"])

    def _stop_container(self, client, cid):
        try:
            client.stop(cid)
        except requests.exceptions.Timeout as e:
            self.logger.warning(f"Failed to stop container {cid}: {str(e)}")
        except docker.errors.APIError as e:
            self.logger.warning(f"Error stopping {cid}: {str(e)}")

    def _build_container_matcher(self, prefixes):

        def matcher(name):
            return any(name.startswith(prefix) for prefix in prefixes)

        return matcher

    def _has_been_running_since(self, container, min_time):
        started_at = container.get("State", {}).get("StartedAt")
        if not started_at:
            return False

        return dateutil.parser.parse(started_at) <= min_time

    def _get_docker_client(self):
        config = self.config.config
        return docker.APIClient(version="auto", timeout=config["http_timeout"])

    def run(self):
        """Autostop main method."""
        self.logger.info("Start autostop")
        config = self.config.config

        if config["stop"]["max_run_time"]:
            self.stop_containers()

        if not config["stop"]["max_run_time"]:
            self.logger.warning("Skipped, no arguments given")



"""Global pytest fixtures."""
import datetime

import pytest
from dateutil import tz


@pytest.fixture
def container():
    return {
        "Id": "abcdabcdabcdabcd",
        "Created": "2013-12-20T17:00:00Z",
        "Name": "/container_name",
        "State": {
            "Running": False,
            "FinishedAt": "2014-01-01T17:30:00Z",
            "StartedAt": "2014-01-01T17:01:00Z",
        }
    }


@pytest.fixture
def containers():
    return [
        {
            "Id": "abcd",
            "Name": "one",
            "Created": "2014-01-01T01:01:01Z",
            "State": {
                "Running": False,
                "FinishedAt": "2014-01-01T01:01:01Z",
            },
        },
        {
            "Id": "abbb",
            "Name": "two",
            "Created": "2014-01-01T01:01:01Z",
            "State": {
                "Running": True,
                "FinishedAt": "2014-01-01T01:01:01Z",
            },
        },
    ]


@pytest.fixture
def image():
    return {
        "Id": "abcdabcdabcdabcd",
        "Created": "2014-01-20T05:00:00Z",
    }


@pytest.fixture
def images():
    return [
        {
            "RepoTags": ["<none>:<none>"],
            "Id": "2471708c19beabababab"
        },
        {
            "RepoTags": ["<none>:<none>"],
            "Id": "babababababaabababab"
        },
        {
            "RepoTags": ["user/one:latest", "user/one:abcd"]
        },
        {
            "RepoTags": ["other-1:abcda"]
        },
        {
            "RepoTags": ["other-2:abc45"]
        },
        {
            "RepoTags": ["new_image:latest", "new_image:123"]
        },
    ]


@pytest.fixture
def now():
    return datetime.datetime(2014, 1, 20, 10, 10, tzinfo=tz.tzutc())


@pytest.fixture
def earlier_time():
    return datetime.datetime(2014, 1, 1, 0, 0, tzinfo=tz.tzutc())


@pytest.fixture
def later_time():
    return datetime.datetime(2014, 1, 20, 0, 10, tzinfo=tz.tzutc())



#!/usr/bin/env python3
"""Global settings definition."""

import os

import anyconfig
import environs
import jsonschema.exceptions
import ruamel.yaml
from appdirs import AppDirs
from jsonschema._utils import format_as_index

import dockertidy.exception
import dockertidy.parser
from dockertidy.parser import env
from dockertidy.utils import Singleton, dict_intersect

config_dir = AppDirs("docker-tidy").user_config_dir
default_config_file = os.path.join(config_dir, "config.yml")


class Config():
    """
    Create an object with all necessary settings.

    Settings are loade from multiple locations in defined order (last wins):
        - default settings defined by `self._get_defaults()`
        - yaml config file, defaults to OS specific user config directory
        - provided cli parameters
    """

    SETTINGS = {
        "command": {
            "default": "",
        },
        "config_file": {
            "default": "",
            "env": "CONFIG_FILE",
            "type": environs.Env().str
        },
        "dry_run": {
            "default": False,
            "env": "DRY_RUN",
            "file": True,
            "type": environs.Env().bool
        },
        "http_timeout": {
            "default": 60,
            "env": "HTTP_TIMEOUT",
            "file": True,
            "type": environs.Env().int
        },
        "logging.level": {
            "default": "WARNING",
            "env": "LOG_LEVEL",
            "file": True,
            "type": environs.Env().str
        },
        "logging.json": {
            "default": False,
            "env": "LOG_JSON",
            "file": True,
            "type": environs.Env().bool
        },
        "gc.max_container_age": {
            "default": "",
            "env": "GC_MAX_CONTAINER_AGE",
            "file": True,
            "type": env.timedelta_validator
        },
        "gc.max_image_age": {
            "default": "",
            "env": "GC_MAX_IMAGE_AGE",
            "file": True,
            "type": env.timedelta_validator
        },
        "gc.dangling_volumes": {
            "default": False,
            "env": "GC_DANGLING_VOLUMES",
            "file": True,
            "type": environs.Env().bool
        },
        "gc.exclude_images": {
            "default": [],
            "env": "GC_EXCLUDE_IMAGES",
            "file": True,
            "type": environs.Env().list
        },
        "gc.exclude_container_labels": {
            "default": [],
            "env": "GC_EXCLUDE_CONTAINER_LABELS",
            "file": True,
            "type": environs.Env().list
        },
        "stop.max_run_time": {
            "default": "",
            "env": "STOP_MAX_RUN_TIME",
            "file": True,
            "type": env.timedelta_validator
        },
        "stop.prefix": {
            "default": [],
            "env": "STOP_PREFIX",
            "file": True,
            "type": environs.Env().list
        },
    }

    def __init__(self, args=None):
        """
        Initialize a new settings class.

        :param args: An optional dict of options, arguments and commands from the CLI.
        :param config_file: An optional path to a yaml config file.
        :returns: None

        """
        if args is None:
            self._args = {}
        else:
            self._args = args
        self._schema = None
        self.config_file = default_config_file
        self.config = None
        self._set_config()

    def _get_args(self, args):
        cleaned = dict(filter(lambda item: item[1] is not None, args.items()))

        normalized = {}
        for key, value in cleaned.items():
            normalized = self._add_dict_branch(normalized, key.split("."), value)

        # Override correct log level from argparse
        levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        log_level = levels.index(self.SETTINGS["logging.level"]["default"])
        if normalized.get("logging"):
            for adjustment in normalized["logging"]["level"]:
                log_level = min(len(levels) - 1, max(log_level + adjustment, 0))
            normalized["logging"]["level"] = levels[log_level]

        return normalized

    def _get_defaults(self, files=False):
        normalized = {}

        for key, item in self.SETTINGS.items():
            if files and not item.get("file"):
                continue
            normalized = self._add_dict_branch(normalized, key.split("."), item["default"])

        self.schema = anyconfig.gen_schema(normalized)
        return normalized

    def _get_envs(self):
        normalized = {}
        for key, item in self.SETTINGS.items():
            if item.get("env"):
                prefix = "TIDY_"
                envname = prefix + item["env"]
                try:
                    value = item["type"](envname)
                    normalized = self._add_dict_branch(normalized, key.split("."), value)
                except environs.EnvError as e:
                    if f'"{envname}" not set' in str(e):
                        pass
                    else:
                        raise dockertidy.exception.ConfigError(
                            "Unable to read environment variable", str(e)
                        ) from e

        return normalized

    def _set_config(self):
        args = self._get_args(self._args)
        envs = self._get_envs()
        defaults = self._get_defaults()
        files_raw = self._get_defaults(files=True)

        # preset config file path
        if envs.get("config_file"):
            self.config_file = self._normalize_path(envs.get("config_file"))

        if args.get("config_file"):
            self.config_file = self._normalize_path(args.get("config_file"))

        source_files = []
        source_files.append(self.config_file)
        source_files.append(os.path.join(os.getcwd(), ".dockertidy"))
        source_files.append(os.path.join(os.getcwd(), ".dockertidy.yml"))
        source_files.append(os.path.join(os.getcwd(), ".dockertidy.yaml"))

        for config in [i for i in source_files if os.path.exists(i)]:
            with open(config, encoding="utf8") as stream:
                s = stream.read()
                try:
                    normalized = ruamel.yaml.safe_load(s)
                except (ruamel.yaml.composer.ComposerError, ruamel.yaml.scanner.ScannerError) as e:
                    message = f"{e.context} {e.problem}"
                    raise dockertidy.exception.ConfigError(
                        f"Unable to read config file {config}", message
                    ) from e

                if self._validate(normalized):
                    anyconfig.merge(files_raw, normalized, ac_merge=anyconfig.MS_DICTS)
                    files_raw["logging"]["level"] = files_raw["logging"]["level"].upper()

        files = dict_intersect(files_raw, self._get_defaults(files=True))

        if self._validate(files):
            anyconfig.merge(defaults, files, ac_merge=anyconfig.MS_DICTS)

        if self._validate(envs):
            anyconfig.merge(defaults, envs, ac_merge=anyconfig.MS_DICTS)

        if self._validate(args):
            anyconfig.merge(defaults, args, ac_merge=anyconfig.MS_DICTS)

        if "config_file" in defaults:
            defaults.pop("config_file")

        defaults["logging"]["level"] = defaults["logging"]["level"].upper()

        self.config = defaults

    def _normalize_path(self, path):
        if not os.path.isabs(path):
            base = os.path.join(os.getcwd(), path)
            return os.path.abspath(os.path.expanduser(os.path.expandvars(base)))

        return path

    def _validate(self, config):
        try:
            anyconfig.validate(config, self.schema, ac_schema_safe=False)
        except jsonschema.exceptions.ValidationError as e:
            schema_error = "Failed validating '{validator}' in schema{schema}\n{message}".format(
                validator=e.validator,
                schema=format_as_index(list(e.relative_schema_path)[:-1]),
                message=e.message
            )
            raise dockertidy.exception.ConfigError("Configuration error", schema_error) from e

        return True

    def _add_dict_branch(self, tree, vector, value):
        key = vector[0]
        tree[key] = value \
            if len(vector) == 1 \
            else self._add_dict_branch(tree[key] if key in tree else {}, vector[1:], value)
        return tree


class SingleConfig(Config, metaclass=Singleton):
    """Singleton config object."""

    pass
